# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Tuple

from rdflib import URIRef

from fhirtordf.rdfsupport.fhirresourcere import FHIR_RESOURCE_RE, FHIR_RE_ID, FHIR_RE_BASE, FHIR_RE_RESOURCE


def uri_to_ide_and_source(uri: URIRef, include_resource: bool=False) -> Tuple[str, str]:
    """
    Convert a  URI into a identifier/ identifier source tuple
    :param uri:  URI
    :param include_resource: If true, resource type becomes part of the identifier
    :return: ide, ide_source
    """
    uri_str = str(uri)
    m = FHIR_RESOURCE_RE.match(uri_str)
    if m:
        ide = ((m.group(FHIR_RE_RESOURCE) + '/') if include_resource else '') + m.group(FHIR_RE_ID)
        ide_source = m.group(FHIR_RE_BASE)
    else:
        # Assume no history entry if not FHIR format...
        ide_source, ide = uri_str.rsplit('#', 1) if '#' in uri_str \
            else uri_str.rsplit('/', 1) if '/' in uri_str else ('UNKNOWN', uri_str)
    return ide, ide_source



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import pickle
import uuid
from typing import Tuple, Optional, Dict


class _PickleJarFactory:
    _instance = None
    _cache_directory = os.path.abspath(os.path.join(os.path.expanduser('~'), ".cache"))

    def clear(self):
        if self._instance:
            self._instance.clear()
        self._instance = None

    @property
    def picklejar(self) -> "_PickleJar":
        if not self._instance:
            self._instance = _PickleJar(self)
        return self._instance

    @property
    def cache_directory(self) -> str:
        return self._cache_directory

    @cache_directory.setter
    def cache_directory(self, path: str) -> None:
        self._cache_directory = path
        self._instance = None

    @property
    def cache_directory_index(self) -> str:
        return os.path.join(self._cache_directory, 'index') if self._cache_directory is not None else None

picklejarfactory = _PickleJarFactory()


class _PickleJar:

    class CacheEntry:
        def __init__(self, sig: Tuple, loc: str):
            """
            An entry in a cache
            :param sig: Unique signature for the entry
            :param loc: Location of the file containing the pickled image of the entry
            """
            self.sig = sig
            self.loc = loc

    def __init__(self, factory: type(_PickleJarFactory)):
        """
        Create a new image cache
        """
        self._cache_directory = factory.cache_directory
        self._cache_directory_index = factory.cache_directory_index
        if self._cache_directory is not None:
            os.makedirs(self._cache_directory, exist_ok=True)
        if self._cache_directory is not None and os.path.exists(self._cache_directory_index):
            self._load()
        else:
            self._cache = {}            # type: Dict[str, _PickleJar.CacheEntry]
            self._update()

    def add(self, name: str, sig: Tuple, obj: object) -> None:
        """
        Add a file to the cache
        :param name: name of the object to be pickled
        :param sig: signature for object
        :param obj: object to pickle
        """
        if self._cache_directory is not None:
            if name in self._cache:
                os.remove(os.path.join(self._cache_directory, self._cache[name].loc))
            fname = os.path.join(self._cache_directory, str(uuid.uuid4()))
            with open(fname, 'wb') as f:
                pickle.dump(obj, f)
            self._cache[name] = _PickleJar.CacheEntry(sig, fname)
            self._update()

    def get(self, name: str, sig: Tuple) -> Optional[object]:
        """
        Return the object representing name if it is cached
        :param name: name of object
        :param sig: unique signature of object
        :return: object if exists and signature matches
        """
        if name not in self._cache:
            return None
        if self._cache[name].sig != sig:
            del self._cache[name]
            self._update()
            return None
        with open(self._cache[name].loc, 'rb') as f:
            return pickle.load(f)

    def _update(self) -> None:
        if self._cache_directory is not None:
            with open(self._cache_directory_index, 'wb') as f:
                pickle.dump(self._cache, f)

    def _load(self) -> None:
        if self._cache_directory is not None:
            with open(self._cache_directory_index, 'rb') as f:
                self._cache = pickle.load(f)
        else:
            self._cache = {}

    def clear(self) -> None:
        """
        Clear all cache entries for directory and, if it is a 'pure' directory, remove the directory itself
        """
        if self._cache_directory is not None:
            # Safety - if there isn't a cache directory file, this probably isn't a valid cache
            assert os.path.exists(self._cache_directory_index), "Attempt to clear a non-existent cache"
            self._load()            # Shouldn't have any impact but...
            for e in self._cache.values():
                if os.path.exists(e.loc):
                    os.remove(e.loc)
            self._cache.clear()
            self._update()
        self._cache = {}


def picklejar():
    return picklejarfactory.picklejar



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional

from rdflib import Namespace, URIRef


class NumericNamespace(Namespace):
    """
    An RDF namespace that supports numeric identifiers (e.g. sct:74400008).
    Notation for use: SCT.C74400008
    """
    def __new__(cls, value):
        return Namespace.__new__(cls, value)

    def __getattr__(self, item: str) -> "URIRef":
        return URIRef(str(self) + item[1:])

    def __eq__(self, other):
        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional, List

from jsonasobj import JsonObj, load
from rdflib import Graph

from fhirtordf.loaders.fhirresourceloader import FHIRResource


class FHIRCollection:
    """ FHIR JSON collection to RDF conversion utility.  This tool takes a collection of json "entry" elements
     and generates a list of FHIRResource elements from the entries in the collection.  The JSON file itself can have
     an optional collection header.
     """
    def __init__(self, vocabulary: Graph, json_fname: Optional[str], base_uri: str, data: Optional[JsonObj] = None,
                 add_ontology_header: Optional[bool] = True, replace_narrative_text: Optional[bool] = False,
                 target: Optional[Graph] = None):
        """
        Convert a JSON collection into RDF.
        :param vocabulary: fhir metadata vocabulary
        :param json_fname: name or URI of the FHIR json collection to convert
        :param base_uri: URI to use as a base for identifiers
        :param data: JsonObj to use if json fname is not present
        :param add_ontology_header: Include the OWL:Ontology declaration
        :param replace_narrative_text: Replace long narrative text with REPLACED_NARRATIVE_TEXT
        :param target: Target graph -- load everything into this if present
        """
        collection = load(json_fname) if json_fname else data

        self.entries = []           # type: List[FHIRResource]
        for entry in collection.entry:
            if 'resource' in entry:
                self.entries.append(FHIRResource(vocabulary, None, base_uri, data=entry.resource,
                                                 add_ontology_header=add_ontology_header,
                                                 replace_narrative_text=replace_narrative_text, target=target))



from distutils.core import setup

import sys

from fhirtordf import __version__

requires = ['python_dateutil', 'rdflib>=4.2.2', 'jsonasobj>=1.1.1', 'dirlistproc>=1.4.5', 'rdflib-jsonld',
            'isodate']
if sys.version_info < (3, 5):
    requires.append('typing')

setup(
    name='fhirtordf',
    version=__version__,
    packages=['tests', 'scripts', 'fhirtordf', 'fhirtordf.fhir', 'fhirtordf.loaders', 'fhirtordf.rdfsupport'],
    url='https://github.com/BD2KOnFHIR/fhirtordf',
    license='Apache 2.0',
    author='Harold Solbrig',
    author_email='solbrig.harold@mayo.edu',
    description='FHIR JSON to RDF Conversion Tool',
    long_description='Convert JSON representatin of FHIR resources into RDF',
    install_requires=requires,
    scripts=['scripts/fhirtordf'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Healthcare Industry',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Database',
        'Programming Language :: Python :: 3 :: Only'
    ]
)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import re
from rdflib import Graph, BNode, Literal
from rdflib.plugins.serializers.turtle import TurtleSerializer, SUBJECT, VERB, _GEN_QNAME_FOR_DT
from rdflib.namespace import NAME_START_CATEGORIES, RDF


# Two changes to the serializer -- one is in the label routine where the serializer (deliberately) mangles the
# format. (use_plain=True).  The second is the actual layout...

class TurtleSeriailizerMod(TurtleSerializer):
    """
    TurtleSerializerMod is an override to the basic TurtleSerializer.

    This change has been tested against rdflib 4.2.2
    other versions may or may not work.
    """
    def label(self, node, position):
        if node == RDF.nil:
            return '()'
        if position is VERB and node in self.keywords:
            return self.keywords[node]
        if isinstance(node, Literal):
            return node._literal_n3(
                # use_plain=True,
                use_plain=False,
                qname_callback=lambda dt: self.getQName(
                    dt, _GEN_QNAME_FOR_DT))
        else:
            node = self.relativize(node)

            return self.getQName(node, position == VERB) or node.n3()

    def p_squared(self, node, position, newline=False):
        if (not isinstance(node, BNode) or node in self._serialized or
                self._references[node] > 1 or position == SUBJECT):
            return False

        if not newline:
            self.write(' ')

        if self.isValidList(node):
            # this is a list
            self.write('(')
            self.depth += 1  # 2
            self.doList(node)
            self.depth -= 1  # 2
            self.write(' )')
        else:
            self.subjectDone(node)
            self.depth += 1
            self.write('[\n' + self.indent())
            # self.write('[')
            self.depth -= 1
            self.predicateList(node, newline=True)
            # self.predicateList(node, newline=False)
            self.write('\n' + self.indent() + ']')
            # self.write(' ]')
            # self.depth -= 1

        return True
TurtleSerializer.p_squared = TurtleSeriailizerMod.p_squared
TurtleSerializer.label = TurtleSeriailizerMod.label

# The following change allows numeric concept identifiers in turtle output.  It is necessary for coding schemes
# such as SNOMED CT.
NAME_START_CATEGORIES.append('Nd')


class PrettyGraph(Graph):
    @staticmethod
    def strip_prefixes(g: Graph):
        """ Remove the prefixes from the graph for aesthetics """
        return re.sub(r'^@prefix .* .\n', '',
                      g.serialize(format="turtle").decode(),
                      flags=re.MULTILINE).strip()

    def serialize(self, format='turtle', **args):
        assert len(args) == 0, "All arguments to serialize are fixed"

        return super().serialize(format=format, spacious=False)
